from enum import Enum
from .request_handler import RequestHandler
from .seldon_http import SeldonRequestHandler
from .tensorflow_http import TensorflowRequestHandler
from .v2 import KFservingV2RequestHandler


class Protocol(Enum):
    tensorflow_http = "tensorflow.http"
    seldon_http = "seldon.http"
    kfserving_http = "kfserving.http"

    def __str__(self):
        return self.value


def get_request_handler(protocol, request: dict) -> RequestHandler:
    """
    Create a request handler for the data

    Parameters
    ----------
    protocol
         Protocol to use
    request
         The incoming request
    Returns
    -------
         A Request Handler for the desired protocol

    """
    if protocol == Protocol.tensorflow_http:
        return TensorflowRequestHandler(request)
    elif protocol == Protocol.seldon_http:
        return SeldonRequestHandler(request)
    elif protocol == Protocol.kfserving_http:
        return KFservingV2RequestHandler(request)
    else:
        raise Exception(f"Unknown protocol {protocol}")



import numpy as np


class RequestHandler(object):
    def __init__(self, request: dict):
        self.request = request

    def validate(self):
        """
        Validate the request

        """
        raise NotImplementedError

    def extract_request(self) -> np.array:
        """
        Extract the request

        Returns
        -------
             A list
        """
        raise NotImplementedError



from .request_handler import RequestHandler
from mlserver.errors import InferenceError
from mlserver.codecs import NumpyCodec
from mlserver import types
import numpy as np


class KFservingV2RequestHandler(RequestHandler):
    def __init__(self, request: dict):
        super().__init__(request)

    def validate(self):
        if "inputs" not in self.request:
            raise InferenceError("Expected key `inputs` in request body")
        # assumes single input
        inputs = self.request["inputs"][0]
        data_type = inputs["datatype"]

        if data_type == "BYTES":
            raise InferenceError(
                "KFServing protocol BYTES data can not be presently handled"
            )

    def extract_request(self) -> np.array:
        inputs = self.request["inputs"][0]
        default_codec = NumpyCodec()
        return default_codec.decode(types.RequestInput(**inputs))



from mlserver import types
from mlserver.settings import ModelSettings
from mlserver.model import MLModel
from mlserver.errors import InferenceError
from mlserver.codecs import NumpyCodec
from fastapi import Request, Response
from mlserver.handlers import custom_handler
from .protocols.util import (
    get_request_handler,
    Protocol,
)
import numpy as np
import orjson
from typing import Optional, Any
from pydantic import BaseSettings, PyObject

ENV_PREFIX_ALIBI_DETECT_SETTINGS = "MLSERVER_MODEL_ALIBI_DETECT_"


class AlibiDetectSettings(BaseSettings):
    """
    Parameters that apply only to alibi detect models
    """

    class Config:
        env_prefix = ENV_PREFIX_ALIBI_DETECT_SETTINGS

    init_detector: bool = False
    detector_type: PyObject = ""
    protocol: Optional[str] = "seldon.http"
    init_parameters: Optional[dict] = {}
    predict_parameters: Optional[dict] = {}


class AlibiDetectParameters(BaseSettings):
    """
    Parameters that apply only to alibi detect models
    """

    predict_parameters: Optional[dict] = {}


class AlibiDetectRuntime(MLModel):
    """
    Implementation of the MLModel interface to load and serve `alibi-detect` models.
    """

    def __init__(self, settings: ModelSettings):

        self.alibi_detect_settings = AlibiDetectSettings(**settings.parameters.extra)
        super().__init__(settings)

    @custom_handler(rest_path="/")
    async def detect(self, request: Request) -> Response:
        """
        This custom handler is meant to mimic the behaviour prediction in alibi-detect
        """
        raw_data = await request.body()
        as_str = raw_data.decode("utf-8")

        try:
            body = orjson.loads(as_str)
        except orjson.JSONDecodeError as e:
            raise InferenceError("Unrecognized request format: %s" % e)

        request_handler = get_request_handler(
            Protocol(self.alibi_detect_settings.protocol), body
        )
        request_handler.validate()
        input_data = request_handler.extract_request()

        y = await self.predict_fn(input_data)
        output_data = orjson.dumps(y, option=orjson.OPT_SERIALIZE_NUMPY)

        return Response(content=output_data, media_type="application/json")

    async def predict(self, payload: types.InferenceRequest) -> types.InferenceResponse:
        payload = self._check_request(payload)
        model_input = payload.inputs[0]
        default_codec = NumpyCodec()
        input_data = self.decode(model_input, default_codec=default_codec)
        y = await self.predict_fn(input_data)

        # TODO: Convert alibi-detect output to v2 protocol
        output_data = np.array(y["data"])

        return types.InferenceResponse(
            model_name=self.name,
            model_version=self.version,
            parameters=y["meta"],
            outputs=[default_codec.encode(name="detect", payload=output_data)],
        )

    async def predict_fn(self, input_data: Any) -> dict:
        parameters = self.alibi_detect_settings.predict_parameters
        return self._model.predict(input_data, **parameters)

    def _check_request(self, payload: types.InferenceRequest) -> types.InferenceRequest:
        if len(payload.inputs) != 1:
            raise InferenceError(
                "AlibiDetector only supports a single input tensor "
                f"({len(payload.inputs)} were received)"
            )
        return payload



from .request_handler import RequestHandler
from mlserver.errors import InferenceError
import numpy as np
from enum import Enum


class SeldonPayload(Enum):
    TENSOR = 1
    NDARRAY = 2
    TFTENSOR = 3


def _extract_list(body: dict) -> np.array:
    data_def = body["data"]
    if "tensor" in data_def:
        arr = np.array(data_def.get("tensor").get("values")).reshape(
            data_def.get("tensor").get("shape")
        )
        return arr
    elif "ndarray" in data_def:
        return np.array(data_def.get("ndarray"))
    elif "tftensor" in data_def:
        arr = np.array(data_def["tftensor"]["float_val"])
        shape = []
        for dim in data_def["tftensor"]["tensor_shape"]["dim"]:
            shape.append(dim["size"])
        arr = arr.reshape(shape)
        return arr
    else:
        raise InferenceError("Unknown Seldon payload %s" % body)


def _get_request_ty(request: dict) -> SeldonPayload:
    data_def = request["data"]
    if "tensor" in data_def:
        return SeldonPayload.TENSOR
    elif "ndarray" in data_def:
        return SeldonPayload.NDARRAY
    elif "tftensor" in data_def:
        return SeldonPayload.TFTENSOR
    else:
        raise InferenceError("Unknown Seldon payload %s" % data_def)


class SeldonRequestHandler(RequestHandler):
    def __init__(self, request: dict):
        super().__init__(request)

    def validate(self):
        if "data" not in self.request:
            raise InferenceError("Expected key `data` in request body")

        _get_request_ty(self.request)

    def extract_request(self) -> np.array:
        return _extract_list(self.request)



from .request_handler import RequestHandler
from mlserver.errors import InferenceError
import numpy as np


class TensorflowRequestHandler(RequestHandler):
    def __init__(self, request: dict):
        super().__init__(request)

    def validate(self):
        if "instances" not in self.request:
            raise InferenceError("Expected key `instances` in request body")

    def extract_request(self) -> np.array:
        return np.array(self.request["instances"])



from mlserver_alibi_detect.runtime import AlibiDetectRuntime
from mlserver_alibi_detect.errors import InvalidAlibiDetector
from mlserver.utils import get_model_uri
from alibi_detect.utils.saving import load_detector
import pickle


class AlibiDriftDetectRuntime(AlibiDetectRuntime):
    """
    Implementation of the MLModel interface to load and serve Alibi drift models.
    """

    async def load(self) -> bool:

        model_uri = await get_model_uri(self._settings)

        if not self.alibi_detect_settings.init_detector:
            try:
                self._model = load_detector(model_uri)
            except (
                ValueError,
                FileNotFoundError,
                EOFError,
                pickle.UnpicklingError,
            ) as e:
                raise InvalidAlibiDetector(self._settings.name) from e
        else:
            try:
                detector_data = pickle.load(open(model_uri, "rb"))
                drift_detector = self.alibi_detect_settings.detector_type
            except (FileNotFoundError, EOFError, pickle.UnpicklingError) as e:
                raise InvalidAlibiDetector(self._settings.name) from e

            parameters = self.alibi_detect_settings.init_parameters

            self._model = drift_detector(**detector_data, **parameters)

        self.ready = True
        return self.ready



from mlserver.errors import MLServerError


class InvalidAlibiDetector(MLServerError):
    def __init__(self, model_name: str):
        msg = f"Invalid Alibi Detector type for model {model_name}"
        super().__init__(msg)
